package com.gxzc.zen.common.util

import java.util.concurrent.ConcurrentHashMap
import java.util.concurrent.Semaphore

/**
 * 细粒度key锁
 * 高并发 维护一个ThreadLocal保证每个线程锁定计数一致
 * 可重入 ReentrantLock(同等),因为使用了ConcurrentHashMap
 * 公平锁 (竞态状态 须排队)
 * @author NorthLan
 * @date 2018/8/30
 * @url https://noahlan.com
 */
open class KeyLock<K> {
    private val map = ConcurrentHashMap<K, Semaphore>()
    private val local = object : ThreadLocal<MutableMap<K, LockInfo>>() {
        override fun initialValue(): MutableMap<K, LockInfo> {
            return mutableMapOf()
        }
    }

    /**
     * 锁定key, 其他需要使用此key的线程进入等待,直到释放锁 {@link #unlock(K)} <br>
     * 使用hashcode和equals来判断key是否相同,因此key必须实现{@link #hashCode()}和{@link #equals(Object)}方法
     */
    fun lock(key: K?) {
        if (key == null) {
            return
        }
        val info = local.get()[key]
        if (info == null) {
            val current = Semaphore(1, false).apply {
                acquireUninterruptibly()
            }
            map.put(key, current)?.apply {
                acquireUninterruptibly()
            }
            local.get()[key] = LockInfo(current)
        } else {
            ++info.lockCount
        }
    }

    /**
     * 释放key锁
     */
    fun unlock(key: K?) {
        if (key == null) {
            return
        }
        val info = local.get()[key]
        if (info != null && --info.lockCount == 0) {
            info.current.release()
            map.remove(key, info.current)
            local.get().remove(key)
        }
    }

    /**
     * 锁定多个key
     * 建议在调用此方法前先对keys进行排序,使用相同的锁定顺序,防止死锁发生
     */
    fun lock(keys: Array<K>?) {
        keys?.forEach {
            lock(it)
        }
    }

    /**
     * 释放多个key
     */
    fun unlock(keys: Array<K>?) {
        keys?.forEach {
            unlock(it)
        }
    }

    /**
     * 锁信息
     */
    private class LockInfo(var current: Semaphore) {
        var lockCount: Int = 1
    }
}