package com.gxzc.zen.common.util

import com.baomidou.mybatisplus.extension.plugins.pagination.Page
import javax.servlet.http.HttpServletRequest

/**
 * 分页工具类
 * @author NorthLan
 * @date 2018/3/23
 * @url https://noahlan.com
 */
object PaginationUtil {
    const val KEY_CURRENT = "current"
    const val KEY_PAGESIZE = "pageSize"

    private fun getCurrent(request: HttpServletRequest): Long? {
        return request.getParameter(KEY_CURRENT).toLong()
    }

    private fun getPageSize(request: HttpServletRequest): Long? {
        return request.getParameter(KEY_PAGESIZE)?.toLong()
    }

    fun <T> getRequestPage(request: HttpServletRequest): Page<T>? {
        if (!paginable(request)) {
            return null
        }
        return Page<T>().apply {
            current = getCurrent(request)!!
            size = getPageSize(request)!!
        }
    }

    /**
     * 逻辑分页 获取分页后的list
     * @param current 当前页码
     * @param size 每页容量
     */
    fun <T> logicPagingList(data: MutableList<T>, current: Long, size: Long): MutableList<T> {
        var fromIndex = (current - 1) * size
        if (fromIndex > data.size) {
            fromIndex = data.size.toLong()
        }
        var toIndex = current * size
        if (toIndex > data.size) {
            toIndex = data.size.toLong()
        }
        return data.subList(fromIndex.toInt(), toIndex.toInt())
    }

    /**
     * 逻辑分页 获取分页信息(包含分页后的数据)
     * @param current 当前页码
     * @param size 每页容量
     * @return 返回值参数示例
     * {
     *   "total": 9,
     *   "size": 4,
     *   "pages": 3,
     *   "current": 1,
     *   "records": [
     *       {
     *           "id": 1,
     *           "enable": true,
     *           "key": "gender",
     *           "value": "1",
     *           "label": "男",
     *           "sort": 1
     *       }
     *   ]
     * }
     */
    fun <T> logicPaging(data: MutableList<T>, current: Long, size: Long): Page<T> {
        return Page<T>(current, size).apply {
            records = logicPagingList(data, current, size)
            setTotal(data.size.toLong())
        }
    }

    fun <T> logicPaging(data: MutableList<T>, request: HttpServletRequest): Page<T> {
        return logicPaging(data, getCurrent(request) ?: 1, getPageSize(request) ?: 20)
    }

    /**
     * 分页参数是否合法
     */
    fun paginable(current: Int?, size: Int?): Boolean {
        return current != null && size != null && current > 0 && size > 0
    }

    /**
     * 分页参数是否合法
     */
    fun paginable(request: HttpServletRequest): Boolean {
        return paginable(getCurrent(request)?.toInt(), getPageSize(request)?.toInt())
    }
}